#version 330 core

in vec2 TexCoords;
in vec3 fragNormal;
in vec3 fragPos;

out vec4 color;

uniform sampler2D texture_diffuse1;
uniform samplerCube depthMaps[8];

uniform float far_plane;

uniform vec3 objectColor;

uniform vec3 lightColor;

const int numberOfPointLights = 8;
uniform vec3 lightPositionPoint1;
uniform vec3 lightPositionPoint2;
uniform vec3 lightPositionPoint3;
uniform vec3 lightPositionPoint4;
uniform vec3 lightPositionPoint5;
uniform vec3 lightPositionPoint6;
uniform vec3 lightPositionPoint7;
uniform vec3 lightPositionPoint8;
vec3 lightPointPosition[numberOfPointLights];

uniform vec3 lightPositionSpot;

uniform vec3 viewPosition;
uniform float cutOff;

uniform bool renderSpot; 

// array of offset direction for sampling
vec3 gridSamplingDisk[20] = vec3[]
(
   vec3(1, 1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1, 1,  1), 
   vec3(1, 1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1, 1, -1),
   vec3(1, 1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1, 1,  0),
   vec3(1, 0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1, 0, -1),
   vec3(0, 1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0, 1, -1)
);


vec3 calculateSpotLight(vec3 normal);
vec3 calculatePointLight(vec3 normal);
float ShadowCalculation(vec3 fragPos, vec3 lightPos, samplerCube depthMap);

void main()
{    
	// todo: DO NOT ADD AMBIENT LIGHT IN SPOT LIGH CALCULATION AND DIRECTIONAL LIGHT CALCULATION
	vec3 norm = normalize(fragNormal);

    vec3 result = calculatePointLight(norm);

	if(renderSpot){
		result += calculateSpotLight(norm);
	}

	// for debugging only
	// ShadowCalculation(fragPos, lightPositionPoint1);

	color = vec4(result, 1.0);	
}


vec3 calculateSpotLight(vec3 normal)
{
	vec3 textureColor = texture(texture_diffuse1, TexCoords).rgb;
	vec3 lightDirection = normalize(lightPositionSpot - fragPos);

	float angle = max(dot(lightDirection, normal), 0);
	vec3 diffuse = angle * lightColor * textureColor;

	float ambientStrength = 0.01f;
    float specularStrength = 0.5f;
	float constant = 1.0f;
	float linear = 0.09;
	float quadratic = 0.032;

	float theta = dot(lightDirection, normalize(-viewPosition));

	if(theta > cutOff){
		float angle = max(dot(lightDirection, normal), 0);
		vec3 diffuse = angle * lightColor * textureColor;

		// calculate attentuation
		float distanceFromLight = length(lightPositionSpot - fragPos);
		float attenuation = 1.0f / (constant + linear * distanceFromLight + quadratic * (distanceFromLight * distanceFromLight));

		vec3 viewDir = normalize(viewPosition - fragPos);
		vec3 reflectDir = reflect(-lightDirection, normal);

		float spec = pow(max(dot(viewDir, reflectDir), 0.0), 32);
		vec3 specular = specularStrength * spec * lightColor;

		// now multiply with attentuation
		diffuse *= attenuation;
		specular *= attenuation;

		vec3 ambient = ambientStrength * lightColor  * textureColor;
		vec3 result = (ambient + diffuse + specular) * objectColor;
		return result;
	}else{
		return vec3(ambientStrength * lightColor  * textureColor);
	}
}
vec3 calculatePointLight(vec3 normal)
{
	lightPointPosition[0] = lightPositionPoint1;
	lightPointPosition[1] = lightPositionPoint2;
	lightPointPosition[2] = lightPositionPoint3;
	lightPointPosition[3] = lightPositionPoint4;
	lightPointPosition[4] = lightPositionPoint5;
	lightPointPosition[5] = lightPositionPoint6;
	lightPointPosition[6] = lightPositionPoint7;
	lightPointPosition[7] = lightPositionPoint8;	

	float ambientStrength = 0.01f;
    float specularStrength = 0.75f;
	float constant = 1.0f;
	float linear = 0.09;
	float quadratic = 0.032;

	vec3 textureColor = texture(texture_diffuse1, TexCoords).rgb;
	vec3 viewDir = normalize(viewPosition - fragPos);
	vec3 lightDirection;
	float angle;
	vec3 diffuse;
	float distanceFromLight; 
	float attenuation;
	vec3 reflectDir; 
	float spec; 
	vec3 specular; 
	
	vec3 ambient = ambientStrength * lightColor  * textureColor;
	vec3 result = ambient;

	for (int index = 0; index < numberOfPointLights; index++)
	{
		lightDirection = normalize(lightPointPosition[index] - fragPos);
		angle = max(dot(lightDirection, normal), 0);
		diffuse = angle * lightColor * textureColor;

		// calculate attentuation
		distanceFromLight = length(lightPointPosition[index] - fragPos);
		attenuation = 1.0f / (constant + linear * distanceFromLight + quadratic * (distanceFromLight * distanceFromLight));

		reflectDir = reflect(-lightDirection, normal);

		spec = pow(max(dot(viewDir, reflectDir), 0.0), 32);
		specular = specularStrength * spec * lightColor;

		// now multiply with attentuation
		diffuse *= attenuation;
		specular *= attenuation;

		float shadow = ShadowCalculation(fragPos, lightPointPosition[index], depthMaps[index]);
		result += (ambient + (1.0 - shadow) * (diffuse + specular)) * objectColor;
	}
	return result;
}

float ShadowCalculation(vec3 fragPos, vec3 lightPos, samplerCube depthMap)
{
	 // get vector between fragment position and light position
    vec3 fragToLight = fragPos - lightPos;
    // use the fragment to light vector to sample from the depth map    
    // float closestDepth = texture(depthMap, fragToLight).r;
    // it is currently in linear range between [0,1], let's re-transform it back to original depth value
    // closestDepth *= far_plane;
    // now get current linear depth as the length between the fragment and light position
    float currentDepth = length(fragToLight);
    // shadow /= (samples * samples * samples);
    float shadow = 0.0;
    float bias = 0.15;
    int samples = 20;
    float viewDistance = length(viewPosition - fragPos);
    float diskRadius = (1.0 + (viewDistance / far_plane)) / 25.0;
    for(int i = 0; i < samples; ++i)
    {
        float closestDepth = texture(depthMap, fragToLight + gridSamplingDisk[i] * diskRadius).r;
        closestDepth *= far_plane;   // undo mapping [0;1]
        if(currentDepth - bias > closestDepth)
            shadow += 1.0;
    }
    shadow /= float(samples);
        
    // for debugging only
    // color = vec4(vec3(closestDepth / far_plane), 1.0);    
        
    return shadow;
}
